__version__="4.3.0"
__copyright__="Copyright (c) 2001 - 2021 The SCons Foundation"
__developer__="bdeegan"
__date__="Tue, 16 Nov 2021 19:09:21 +0000"
__buildsys__="octodog"
__revision__="559790274f66fa55251f5754de34820a29c7327a"
__build__="559790274f66fa55251f5754de34820a29c7327a"
# make sure compatibility is always in place
import SCons.compat # noqa